`default_nettype none
module i2ssdclocksynch(data, lrck, bck, sdclock, sdclock5, datasynch, lrcksynch, bcksynch, tel0);
// Bemonstert data, lrck en bck 1 keer per sdclock-periode met de
// vervijfvoudigde klok sdclock5, en probeert de 1/5 sdclock-periode zodanig
// te kiezen dat er geen metastabiliteit optreedt in het resultaat.
// bck moet een frequentie hebben die gelijk is aan die van sdclock gedeeld
// door een geheel getal van 1 tot 200.
// Samples data, lrck and bck once per sdclock period with the five times
// higher clock sdclock5, and tries to choose the 1/5 sdclock period such that
// no metastability occurs in the result.
// bck must have a frequency that equals that of sdclock divided by an integer
// number from 1 to 200.
input wire data, lrck;
(* CLOCK_SIGNAL = "TRUE" *) input wire bck;
(* CLOCK_SIGNAL = "TRUE" *) input wire sdclock;
(* CLOCK_SIGNAL = "TRUE" *) input wire sdclock5;
output reg datasynch = 1'b0;
output reg lrcksynch = 1'b0;
output reg bcksynch = 1'b0;
output reg tel0;
reg bckdiv2 = 1'b0;
reg sdclockdiv2 = 1'b0;
(* ASYNC_REG = "TRUE" *) reg a = 1'b0;
(* ASYNC_REG = "TRUE" *) reg b = 1'b0;
reg c = 1'b0;
wire d;
reg e = 1'b0;
reg f = 1'b0;
reg g = 1'b0;
reg [2:0] h = 3'b000;
reg [2:0] hh = 3'b000;
(* IOB = "TRUE" *) reg data2 = 1'b0;
(* IOB = "TRUE" *) reg lrck2 = 1'b0;
reg [2:0] teller = 3'b000; // teller die bijhoudt in welk deel van een sdclock-cyclus we zitten, counter that keeps track of where we are in an sdclock cycle
reg [10:0] bckteller = 11'b00000000000; // Als je in elke cyclus van sdclock 1 monster van bck neemt, aliast bck naar 0 Hz als je DSD512 afspeelt,
// omdat daarbij de bck- en sdclock-frequenties gelijk zijn. bckteller wordt gebruikt om dit te detecteren zodat in dat geval sdclock 
// gebruikt kan worden als bitklok.
// When you take 1 sample of bck in each sdclock cycle, bck will alias to 0 Hz in case of DSD512, where the bit clock and sdclock frequencies are equal. 
// bckteller is used to detect this, so the sdclock can then be used as bit clock. 
reg gelijk; // uitgang van een vergelijker, wordt gebruikt als clock enable bij het inklokken van bck, data2 en lrck2 in het sdclock5-domein
// comparator output that is used as a clock enable for clocking in bck, data2 and lrck2 to the sdclock5 domain
reg bcksynch1 = 1'b0;
reg bcksynch1d = 1'b0;
reg datasynch1 = 1'b0;
reg lrcksynch1 = 1'b0;

always @ (posedge bck)
begin
	data2 <= data;
	lrck2 <= lrck;
	bckdiv2 <= !bckdiv2;
end
always @ (posedge sdclock5)
begin
	a <= bckdiv2;
	b <= a;
	c <= b;
end
assign d = b ^^ c; // puls na opgaande flank bck, pulse after bck rising edge
always @ (posedge sdclock) sdclockdiv2 <= !sdclockdiv2;
always @ (posedge sdclock5)
begin
	e <= sdclockdiv2;
	f <= e;
	g <= f;
	if (f ^^ g) // paar sdclock5-cycli na stijgende flank sdclock, few cycles after sdclock rising edge
	begin
		teller <= 3'b000;
	end
	else
	begin
		teller <= (teller + 3'b001) % 8;
	end
	if (d == 1'b1) h <= teller; // teller bemonsterd na een stijgende flank op bck, teller sampled after a rising bck edge

	// hh bepaalt bij welke waarde van teller bck, data2 en lrck2
	// bemonsterd worden. Mocht uit vergelijking van hh met de na een
	// stijgende flank van bck bemonsterde waarde van teller 
	// (variabele h) blijken dat dat op een ongeschikt
	// moment gebeurt, dan moet hh worden aangepast.
	// Feitelijk is dit een hysteresedinges: rechtstreeks de bemonsterde
	// waarde van teller gebruiken kan niet omdat die willekeurig tussen 
	// twee waarden heen en weer zou kunnen springen, hh zou daar vrij 
	// van moeten zijn.
	// hh determines at what value of teller bck, data2 and lrck2 get
	// sampled. Should comparing hh to the after a rising edge of bck
	// sampled value of teller (in variable h) show that this happens at
	// an unsuitable moment, then hh must be adjusted.
	// This is actually a hysteretic thing: directly using the sampled
	// value of teller is not possible it might randomly jump between two
	// values, hh should be free of that.
	case ({h, hh})
		6'b000010: hh <= h;
		6'b000011: hh <= h;
		6'b001011: hh <= h;
		6'b001100: hh <= h;
		6'b010000: hh <= h;
		6'b010100: hh <= h;
		6'b011000: hh <= h;
		6'b011001: hh <= h;
		6'b100001: hh <= h;
		6'b100010: hh <= h;
		default: hh <= hh;
	endcase
end
always @ (*)
begin
	if (teller == hh)
		gelijk = 1'b1;
	else
		gelijk = 1'b0;
	tel0 = !(teller[2] || teller[1] || teller[0]);
end
always @ (posedge sdclock5)
begin
	if (gelijk == 1'b1)
	begin
		bcksynch1 <= bck;
		datasynch1 <= data2;
		lrcksynch1 <= lrck2;
	end
	bcksynch1d <= bcksynch1;
	if (tel0 == 1'b1)
	begin
		datasynch <= datasynch1;
		lrcksynch <= lrcksynch1;
	end
	if (bckteller[10]) // dat wil zeggen, als bcksynch1 naar 0 Hz aliast, that is, when bcksynch1 aliases to 0 Hz
		bcksynch <= (!teller[2]) && (!teller[1]);
	else
		if (tel0 == 1'b1) bcksynch <= bcksynch1;
	if (bcksynch1 == bcksynch1d)
	begin
		if (!bckteller[10]) bckteller <= bckteller + 11'b00000000001;
	end
	else
	begin
		bckteller <= 11'b00000000000;
	end
end
endmodule
`default_nettype wire
